@@ -36,11 +36,15 @@
 #include "fsal.h"
 #include "FSAL/fsal_commonlib.h"
 #include "FSAL/fsal_config.h"
+#include "fsal_convert.h"
 #include "config_parsing.h"
 #include "gluster_internal.h"
 #include "nfs_exports.h"
 #include "export_mgr.h"
 #include "pnfs_utils.h"
+#include "mdcache.h"
+#include "sal_data.h"
+
 /* The default location of gfapi log
  * if glfs_log param is not defined in
  * the export file */
@@ -54,8 +58,6 @@ static void export_release(struct fsal_export *exp_hdl)
 {
 	struct glusterfs_export *glfs_export =
 	    container_of(exp_hdl, struct glusterfs_export, export);
-	int *retval = NULL;
-	int err     = 0;
 
 	/* check activity on the export */
 
@@ -64,24 +66,8 @@ static void export_release(struct fsal_export *exp_hdl)
 			   &glfs_export->export.exports);
 	free_export_ops(&glfs_export->export);
 
-	atomic_add_int8_t (&glfs_export->destroy_mode, 1);
+	glusterfs_free_fs(glfs_export->gl_fs);
 
-	/* Wait for up_thread to exit */
-	err = pthread_join(glfs_export->up_thread, (void **)&retval);
-
-	if (retval && *retval) {
-		LogDebug(COMPONENT_FSAL, "Up_thread join returned value %d",
-			 *retval);
-	}
-
-	if (err) {
-		LogCrit(COMPONENT_FSAL, "Up_thread join failed (%s)",
-			strerror(err));
-		return;
-	}
-
-	/* Gluster and memory cleanup */
-	glfs_fini(glfs_export->gl_fs);
 	glfs_export->gl_fs = NULL;
 	gsh_free(glfs_export->export_path);
 	glfs_export->export_path = NULL;
@@ -95,7 +81,8 @@ static void export_release(struct fsal_export *exp_hdl)
 
 static fsal_status_t lookup_path(struct fsal_export *export_pub,
 				 const char *path,
-				 struct fsal_obj_handle **pub_handle)
+				 struct fsal_obj_handle **pub_handle,
+				 struct attrlist *attrs_out)
 {
 	int rc = 0;
 	fsal_status_t status = { ERR_FSAL_NO_ERROR, 0 };
@@ -113,44 +100,33 @@ static fsal_status_t lookup_path(struct fsal_export *export_pub,
 	*pub_handle = NULL;
 
 	if (strcmp(path, glfs_export->mount_path) == 0) {
-		realpath = strdup(glfs_export->export_path);
+		realpath = gsh_strdup(glfs_export->export_path);
 	} else {
 		/*
 		 *  mount path is not same as the exported one. Should be subdir
 		 *  then.
-		 *  TODO: How do we handle symlinks if present in the path.
+		 */
+		/** @todo: How do we handle symlinks if present in the path.
 		 */
 		realpath = gsh_malloc(strlen(glfs_export->export_path) +
 				      strlen(path) + 1);
-		if (realpath) {
-			/*
-			 * Handle the case wherein glfs_export->export_path
-			 * is root i.e, '/' separately.
-			 */
-			if (strlen(glfs_export->export_path) != 1) {
-				strcpy(realpath, glfs_export->export_path);
-				strcpy((realpath +
-					strlen(glfs_export->export_path)),
-					&path[strlen(glfs_export->mount_path)]);
-			} else {
-				strcpy(realpath,
-					&path[strlen(glfs_export->mount_path)]);
-			}
+		/*
+		 * Handle the case wherein glfs_export->export_path
+		 * is root i.e, '/' separately.
+		 */
+		if (strlen(glfs_export->export_path) != 1) {
+			strcpy(realpath, glfs_export->export_path);
+			strcpy((realpath +
+				strlen(glfs_export->export_path)),
+				&path[strlen(glfs_export->mount_path)]);
+		} else {
+			strcpy(realpath,
+				&path[strlen(glfs_export->mount_path)]);
 		}
 	}
-	if (!realpath) {
-		errno = ENOMEM;
-		status = gluster2fsal_error(errno);
-		goto out;
-	}
 
-#ifdef USE_GLUSTER_SYMLINK_MOUNT
-		glhandle = glfs_h_lookupat(glfs_export->gl_fs, NULL, realpath,
-					&sb, 1);
-#else
-		glhandle = glfs_h_lookupat(glfs_export->gl_fs, NULL, realpath,
-					&sb);
-#endif
+	glhandle = glfs_h_lookupat(glfs_export->gl_fs->fs, NULL, realpath,
+				&sb, 1);
 	if (glhandle == NULL) {
 		status = gluster2fsal_error(errno);
 		goto out;
@@ -162,29 +138,28 @@ static fsal_status_t lookup_path(struct fsal_export *export_pub,
 		goto out;
 	}
 
-	rc = glfs_get_volumeid(glfs_export->gl_fs, vol_uuid, GLAPI_UUID_LENGTH);
+	rc = glfs_get_volumeid(glfs_export->gl_fs->fs, vol_uuid,
+			       GLAPI_UUID_LENGTH);
 	if (rc < 0) {
 		status = gluster2fsal_error(rc);
 		goto out;
 	}
 
-	rc = construct_handle(glfs_export, &sb, glhandle, globjhdl,
-			      GLAPI_HANDLE_LENGTH, &objhandle, vol_uuid);
-	if (rc != 0) {
-		status = gluster2fsal_error(rc);
-		goto out;
+	construct_handle(glfs_export, &sb, glhandle, globjhdl,
+			 GLAPI_HANDLE_LENGTH, &objhandle, vol_uuid);
+
+	if (attrs_out != NULL) {
+		posix2fsal_attributes(&sb, attrs_out);
 	}
 
 	*pub_handle = &objhandle->handle;
 
-	if (realpath)
-		free(realpath);
+	gsh_free(realpath);
 
 	return status;
  out:
 	gluster_cleanup_vars(glhandle);
-	if (realpath)
-		free(realpath);
+	gsh_free(realpath);
 
 	return status;
 }
@@ -232,7 +207,8 @@ static fsal_status_t extract_handle(struct fsal_export *exp_hdl,
 
 static fsal_status_t create_handle(struct fsal_export *export_pub,
 				   struct gsh_buffdesc *fh_desc,
-				   struct fsal_obj_handle **pub_handle)
+				   struct fsal_obj_handle **pub_handle,
+				   struct attrlist *attrs_out)
 {
 	int rc = 0;
 	fsal_status_t status = { ERR_FSAL_NO_ERROR, 0 };
@@ -261,24 +237,25 @@ static fsal_status_t create_handle(struct fsal_export *export_pub,
 	memcpy(globjhdl, fh_desc->addr+GLAPI_UUID_LENGTH, GFAPI_HANDLE_LENGTH);
 
 	glhandle =
-	    glfs_h_create_from_handle(glfs_export->gl_fs, globjhdl,
+	    glfs_h_create_from_handle(glfs_export->gl_fs->fs, globjhdl,
 				      GFAPI_HANDLE_LENGTH, &sb);
 	if (glhandle == NULL) {
 		status = gluster2fsal_error(errno);
 		goto out;
 	}
 
-	rc = glfs_get_volumeid(glfs_export->gl_fs, vol_uuid, GLAPI_UUID_LENGTH);
+	rc = glfs_get_volumeid(glfs_export->gl_fs->fs, vol_uuid,
+			       GLAPI_UUID_LENGTH);
 	if (rc < 0) {
 		status = gluster2fsal_error(rc);
 		goto out;
 	}
 
-	rc = construct_handle(glfs_export, &sb, glhandle, globjhdl,
-			      GLAPI_HANDLE_LENGTH, &objhandle, vol_uuid);
-	if (rc != 0) {
-		status = gluster2fsal_error(rc);
-		goto out;
+	construct_handle(glfs_export, &sb, glhandle, globjhdl,
+			 GLAPI_HANDLE_LENGTH, &objhandle, vol_uuid);
+
+	if (attrs_out != NULL) {
+		posix2fsal_attributes(&sb, attrs_out);
 	}
 
 	*pub_handle = &objhandle->handle;
@@ -306,7 +283,8 @@ static fsal_status_t get_dynamic_info(struct fsal_export *exp_hdl,
 	struct glusterfs_export *glfs_export =
 	    container_of(exp_hdl, struct glusterfs_export, export);
 
-	rc = glfs_statvfs(glfs_export->gl_fs, glfs_export->export_path, &vfssb);
+	rc = glfs_statvfs(glfs_export->gl_fs->fs, glfs_export->export_path,
+			  &vfssb);
 	if (rc != 0)
 		return gluster2fsal_error(rc);
 
@@ -323,8 +301,40 @@ static fsal_status_t get_dynamic_info(struct fsal_export *exp_hdl,
 	return status;
 }
 
+/**
+ * @brief Allocate a state_t structure
+ *
+ * Note that this is not expected to fail since memory allocation is
+ * expected to abort on failure.
+ *
+ * @param[in] exp_hdl               Export state_t will be associated with
+ * @param[in] state_type            Type of state to allocate
+ * @param[in] related_state         Related state if appropriate
+ *
+ * @returns a state structure.
+ */
+
+struct state_t *glusterfs_alloc_state(struct fsal_export *exp_hdl,
+				enum state_type state_type,
+				struct state_t *related_state)
+{
+	struct state_t *state;
+	struct glusterfs_fd *my_fd;
+
+	state = init_state(gsh_calloc(1, sizeof(struct state_t)
+					 + sizeof(struct glusterfs_fd)),
+			   exp_hdl, state_type, related_state);
+
+	my_fd = (struct glusterfs_fd *)(state + 1);
+
+	my_fd->glfd = NULL;
+
+	return state;
+}
+
 /** @todo: We have gone POSIX way for the APIs below, can consider the CEPH way
- * in case all are constants across all volumes etc. */
+ * in case all are constants across all volumes etc.
+ */
 
 /**
  * @brief Implements GLUSTER FSAL exportoperation fs_supports
@@ -442,9 +452,15 @@ static fsal_aclsupp_t fs_acl_support(struct fsal_export *exp_hdl)
 static attrmask_t fs_supported_attrs(struct fsal_export *exp_hdl)
 {
 	struct fsal_staticfsinfo_t *info;
+	struct glusterfs_export *glfs_export =
+	    container_of(exp_hdl, struct glusterfs_export, export);
+	attrmask_t supported_mask;
 
 	info = gluster_staticinfo(exp_hdl->fsal);
-	return fsal_supported_attrs(info);
+	supported_mask = fsal_supported_attrs(info);
+	if (!NFSv4_ACL_SUPPORT)
+		supported_mask &= ~ATTR_ACL;
+	return supported_mask;
 }
 
 /**
@@ -539,6 +555,7 @@ void export_ops_init(struct export_ops *ops)
 	ops->fs_supported_attrs = fs_supported_attrs;
 	ops->fs_umask = fs_umask;
 	ops->fs_xattr_access_rights = fs_xattr_access_rights;
+	ops->alloc_state = glusterfs_alloc_state;
 }
 
 struct glexport_params {
@@ -571,6 +588,154 @@ static struct config_block export_param = {
 	.blk_desc.u.blk.commit = noop_conf_commit
 };
 
+/*
+ * Given glusterfs_fs object, decrement the refcount. In case if it
+ * becomes zero, free the resources.
+ */
+void
+glusterfs_free_fs(struct glusterfs_fs *gl_fs)
+{
+	int64_t refcnt;
+	int *retval = NULL;
+	int err     = 0;
+
+	PTHREAD_MUTEX_lock(&GlusterFS.lock);
+
+	refcnt = --(gl_fs->refcnt);
+
+	assert(refcnt >= 0);
+
+	if (refcnt) {
+		LogDebug(COMPONENT_FSAL,
+			 "There are still (%ld)active shares for volume(%s)",
+			 gl_fs->refcnt, gl_fs->volname);
+		PTHREAD_MUTEX_unlock(&GlusterFS.lock);
+		return;
+	}
+
+	glist_del(&gl_fs->fs_obj);
+	PTHREAD_MUTEX_unlock(&GlusterFS.lock);
+
+	atomic_inc_int8_t(&gl_fs->destroy_mode);
+
+	/* Wait for up_thread to exit */
+	err = pthread_join(gl_fs->up_thread, (void **)&retval);
+
+	if (retval && *retval) {
+		LogDebug(COMPONENT_FSAL, "Up_thread join returned value %d",
+			 *retval);
+	}
+
+	if (err) {
+		LogCrit(COMPONENT_FSAL, "Up_thread join failed (%s)",
+			strerror(err));
+		return;
+	}
+
+	/* Gluster and memory cleanup */
+	glfs_fini(gl_fs->fs);
+	gsh_free(gl_fs->volname);
+	gsh_free(gl_fs);
+}
+
+/**
+ * @brief Given Gluster export params, find and return if there is
+ * already existing export entry. If not create one.
+ */
+struct glusterfs_fs*
+glusterfs_get_fs(struct glexport_params params,
+		 const struct fsal_up_vector *up_ops)
+{
+	int rc = 0;
+	struct glusterfs_fs *gl_fs = NULL;
+	glfs_t  *fs = NULL;
+	struct glist_head *glist, *glistn;
+
+	PTHREAD_MUTEX_lock(&GlusterFS.lock);
+
+	glist_for_each_safe(glist, glistn, &GlusterFS.fs_obj) {
+		gl_fs = glist_entry(glist, struct glusterfs_fs,
+				    fs_obj);
+		if (!strcmp(params.glvolname, gl_fs->volname)) {
+			goto found;
+		}
+	}
+
+	gl_fs = gsh_calloc(1, sizeof(struct glusterfs_fs));
+
+	if (!gl_fs) {
+		LogCrit(COMPONENT_FSAL,
+			"Unable to allocate memory for glusterfs_fs object");
+		goto out;
+	}
+
+	glist_init(&gl_fs->fs_obj);
+
+	fs = glfs_new(params.glvolname);
+	if (!fs) {
+		LogCrit(COMPONENT_FSAL,
+			"Unable to create new glfs. Volume: %s",
+			params.glvolname);
+		goto out;
+	}
+
+	rc = glfs_set_volfile_server(fs, "tcp", params.glhostname, 24007);
+	if (rc != 0) {
+		LogCrit(COMPONENT_FSAL,
+			"Unable to set volume file. Volume: %s",
+			params.glvolname);
+		goto out;
+	}
+
+	rc = glfs_set_logging(fs, params.glfs_log, 7);
+	if (rc != 0) {
+		LogCrit(COMPONENT_FSAL,
+			"Unable to set logging. Volume: %s",
+			params.glvolname);
+		goto out;
+	}
+
+	rc = glfs_init(fs);
+	if (rc != 0) {
+		LogCrit(COMPONENT_FSAL,
+			"Unable to initialize volume. Volume: %s",
+			params.glvolname);
+		goto out;
+	}
+
+	gl_fs->fs = fs;
+	gl_fs->volname = strdup(params.glvolname);
+	gl_fs->destroy_mode = 0;
+
+	gl_fs->up_ops = up_ops;
+	rc = initiate_up_thread(gl_fs);
+	if (rc != 0) {
+		LogCrit(COMPONENT_FSAL,
+			"Unable to create GLUSTERFSAL_UP_Thread. Volume: %s",
+			params.glvolname);
+		goto out;
+	}
+
+	glist_add(&GlusterFS.fs_obj, &gl_fs->fs_obj);
+
+found:
+	++(gl_fs->refcnt);
+	PTHREAD_MUTEX_unlock(&GlusterFS.lock);
+	return gl_fs;
+
+out:
+	PTHREAD_MUTEX_unlock(&GlusterFS.lock);
+	if (fs)
+		glfs_fini(fs);
+
+	if (gl_fs) {
+		glist_del(&gl_fs->fs_obj); /* not needed atm */
+		gsh_free(gl_fs);
+	}
+
+	return NULL;
+}
+
 /**
  * @brief Implements GLUSTER FSAL moduleoperation create_export
  */
@@ -583,7 +748,6 @@ fsal_status_t glusterfs_create_export(struct fsal_module *fsal_hdl,
 	int rc = 0;
 	fsal_status_t status = { ERR_FSAL_NO_ERROR, 0 };
 	struct glusterfs_export *glfsexport = NULL;
-	glfs_t *fs = NULL;
 	struct glexport_params params = {
 		.glvolname = NULL,
 		.glhostname = NULL,
@@ -591,7 +755,7 @@ fsal_status_t glusterfs_create_export(struct fsal_module *fsal_hdl,
 		.glfs_log = NULL};
 
 	LogDebug(COMPONENT_FSAL, "In args: export path = %s",
-		 op_ctx->export->fullpath);
+		 op_ctx->ctx_export->fullpath);
 
 	rc = load_config_from_node(parse_node,
 				   &export_param,
@@ -601,7 +765,7 @@ fsal_status_t glusterfs_create_export(struct fsal_module *fsal_hdl,
 	if (rc != 0) {
 		LogCrit(COMPONENT_FSAL,
 			"Incorrect or missing parameters for export %s",
-			op_ctx->export->fullpath);
+			op_ctx->ctx_export->fullpath);
 		status.major = ERR_FSAL_INVAL;
 		goto out;
 	}
@@ -609,57 +773,13 @@ fsal_status_t glusterfs_create_export(struct fsal_module *fsal_hdl,
 		 params.glvolname, params.glvolpath);
 
 	glfsexport = gsh_calloc(1, sizeof(struct glusterfs_export));
-	if (glfsexport == NULL) {
-		status.major = ERR_FSAL_NOMEM;
-		LogCrit(COMPONENT_FSAL,
-			"Unable to allocate export object.  Export: %s",
-			op_ctx->export->fullpath);
-		goto out;
-	}
-
-	if (fsal_export_init(&glfsexport->export) != 0) {
-		status.major = ERR_FSAL_NOMEM;
-		LogCrit(COMPONENT_FSAL,
-			"Unable to allocate export ops vectors.  Export: %s",
-			op_ctx->export->fullpath);
-		goto out;
-	}
 
+	fsal_export_init(&glfsexport->export);
 	export_ops_init(&glfsexport->export.exp_ops);
 
-	fs = glfs_new(params.glvolname);
-	if (!fs) {
+	glfsexport->gl_fs = glusterfs_get_fs(params, up_ops);
+	if (!glfsexport->gl_fs) {
 		status.major = ERR_FSAL_SERVERFAULT;
-		LogCrit(COMPONENT_FSAL,
-			"Unable to create new glfs. Export: %s",
-			op_ctx->export->fullpath);
-		goto out;
-	}
-
-	rc = glfs_set_volfile_server(fs, "tcp", params.glhostname, 24007);
-	if (rc != 0) {
-		status.major = ERR_FSAL_SERVERFAULT;
-		LogCrit(COMPONENT_FSAL,
-			"Unable to set volume file. Export: %s",
-			op_ctx->export->fullpath);
-		goto out;
-	}
-
-	rc = glfs_set_logging(fs, params.glfs_log, 7);
-	if (rc != 0) {
-		status.major = ERR_FSAL_SERVERFAULT;
-		LogCrit(COMPONENT_FSAL,
-			"Unable to set logging. Export: %s",
-			op_ctx->export->fullpath);
-		goto out;
-	}
-
-	rc = glfs_init(fs);
-	if (rc != 0) {
-		status.major = ERR_FSAL_SERVERFAULT;
-		LogCrit(COMPONENT_FSAL,
-			"Unable to initialize volume. Export: %s",
-			op_ctx->export->fullpath);
 		goto out;
 	}
 
@@ -667,19 +787,17 @@ fsal_status_t glusterfs_create_export(struct fsal_module *fsal_hdl,
 	if (rc != 0) {
 		status.major = ERR_FSAL_SERVERFAULT;
 		LogCrit(COMPONENT_FSAL, "Unable to attach export. Export: %s",
-			op_ctx->export->fullpath);
+			op_ctx->ctx_export->fullpath);
 		goto out;
 	}
 
-	glfsexport->mount_path = op_ctx->export->fullpath;
+	glfsexport->mount_path = op_ctx->ctx_export->fullpath;
 	glfsexport->export_path = params.glvolpath;
-	glfsexport->gl_fs = fs;
 	glfsexport->saveduid = geteuid();
 	glfsexport->savedgid = getegid();
 	glfsexport->export.fsal = fsal_hdl;
 	glfsexport->acl_enable =
-		!(op_ctx->export->options & EXPORT_OPTION_DISABLE_ACL);
-	glfsexport->destroy_mode = 0;
+		!op_ctx_export_has_option(EXPORT_OPTION_DISABLE_ACL);
 
 	op_ctx->fsal_export = &glfsexport->export;
 
@@ -695,8 +813,9 @@ fsal_status_t glusterfs_create_export(struct fsal_module *fsal_hdl,
 			goto out;
 
 		/* special case: server_id matches export_id */
-		pds->id_servers = op_ctx->export->export_id;
-		pds->mds_export = op_ctx->export;
+		pds->id_servers = op_ctx->ctx_export->export_id;
+		pds->mds_export = op_ctx->ctx_export;
+		pds->mds_fsal_export = op_ctx->fsal_export;
 
 		if (!pnfs_ds_insert(pds)) {
 			LogCrit(COMPONENT_CONFIG,
@@ -708,7 +827,7 @@ fsal_status_t glusterfs_create_export(struct fsal_module *fsal_hdl,
 
 		LogDebug(COMPONENT_PNFS,
 			 "glusterfs_fsal_create: pnfs ds was enabled for [%s]",
-			 op_ctx->export->fullpath);
+			 op_ctx->ctx_export->fullpath);
 	}
 
 	glfsexport->pnfs_mds_enabled =
@@ -717,23 +836,12 @@ fsal_status_t glusterfs_create_export(struct fsal_module *fsal_hdl,
 	if (glfsexport->pnfs_mds_enabled) {
 		LogDebug(COMPONENT_PNFS,
 			 "glusterfs_fsal_create: pnfs mds was enabled for [%s]",
-			 op_ctx->export->fullpath);
+			 op_ctx->ctx_export->fullpath);
 		export_ops_pnfs(&glfsexport->export.exp_ops);
 		fsal_ops_pnfs(&glfsexport->export.fsal->m_ops);
 	}
 
-	if (up_ops) {
-		glfsexport->export.up_ops = up_ops;
-		rc = initiate_up_thread(glfsexport);
-
-		if (rc != 0) {
-			LogCrit(COMPONENT_FSAL,
-				"Unable to create GLUSTERFSAL_UP_Thread. Export: %s",
-				op_ctx->export->fullpath);
-			status.major = ERR_FSAL_FAULT;
-			goto out;
-		}
-	}
+	glfsexport->export.up_ops = up_ops;
 
  out:
 	if (params.glvolname)
@@ -747,12 +855,10 @@ fsal_status_t glusterfs_create_export(struct fsal_module *fsal_hdl,
 		if (params.glvolpath)
 			gsh_free(params.glvolpath);
 
-		if (fs)
-			glfs_fini(fs);
-
 		if (glfsexport)
 			gsh_free(glfsexport);
 	}
 
 	return status;
 }
+
